
#ifndef ring_core_generated_PREFIX_SYMBOLS_H
#define ring_core_generated_PREFIX_SYMBOLS_H

#define ecp_nistz256_point_double p256_point_double
#define ecp_nistz256_point_add p256_point_add
#define ecp_nistz256_point_add_affine p256_point_add_affine
#define ecp_nistz256_ord_mul_mont p256_scalar_mul_mont
#define ecp_nistz256_ord_sqr_mont p256_scalar_sqr_rep_mont
#define ecp_nistz256_mul_mont p256_mul_mont
#define ecp_nistz256_sqr_mont p256_sqr_mont
#define adx_bmi2_available ring_core_0_17_14__adx_bmi2_available
#define avx2_available ring_core_0_17_14__avx2_available
#define CRYPTO_memcmp ring_core_0_17_14__CRYPTO_memcmp
#define CRYPTO_poly1305_finish ring_core_0_17_14__CRYPTO_poly1305_finish
#define CRYPTO_poly1305_finish_neon ring_core_0_17_14__CRYPTO_poly1305_finish_neon
#define CRYPTO_poly1305_init ring_core_0_17_14__CRYPTO_poly1305_init
#define CRYPTO_poly1305_init_neon ring_core_0_17_14__CRYPTO_poly1305_init_neon
#define CRYPTO_poly1305_update ring_core_0_17_14__CRYPTO_poly1305_update
#define CRYPTO_poly1305_update_neon ring_core_0_17_14__CRYPTO_poly1305_update_neon
#define ChaCha20_ctr32 ring_core_0_17_14__ChaCha20_ctr32
#define ChaCha20_ctr32_avx2 ring_core_0_17_14__ChaCha20_ctr32_avx2
#define ChaCha20_ctr32_neon ring_core_0_17_14__ChaCha20_ctr32_neon
#define ChaCha20_ctr32_nohw ring_core_0_17_14__ChaCha20_ctr32_nohw
#define ChaCha20_ctr32_ssse3 ring_core_0_17_14__ChaCha20_ctr32_ssse3
#define ChaCha20_ctr32_ssse3_4x ring_core_0_17_14__ChaCha20_ctr32_ssse3_4x
#define LIMB_is_zero ring_core_0_17_14__LIMB_is_zero
#define LIMBS_add_mod ring_core_0_17_14__LIMBS_add_mod
#define LIMBS_are_zero ring_core_0_17_14__LIMBS_are_zero
#define LIMBS_equal ring_core_0_17_14__LIMBS_equal
#define LIMBS_less_than ring_core_0_17_14__LIMBS_less_than
#define LIMBS_reduce_once ring_core_0_17_14__LIMBS_reduce_once
#define LIMBS_select_512_32 ring_core_0_17_14__LIMBS_select_512_32
#define LIMBS_shl_mod ring_core_0_17_14__LIMBS_shl_mod
#define LIMBS_sub_mod ring_core_0_17_14__LIMBS_sub_mod
#define LIMBS_window5_split_window ring_core_0_17_14__LIMBS_window5_split_window
#define LIMBS_window5_unsplit_window ring_core_0_17_14__LIMBS_window5_unsplit_window
#define LIMB_shr ring_core_0_17_14__LIMB_shr
#define OPENSSL_cpuid_setup ring_core_0_17_14__OPENSSL_cpuid_setup
#define aes_gcm_dec_kernel ring_core_0_17_14__aes_gcm_dec_kernel
#define aes_gcm_dec_update_vaes_avx2 ring_core_0_17_14__aes_gcm_dec_update_vaes_avx2
#define aes_gcm_enc_kernel ring_core_0_17_14__aes_gcm_enc_kernel
#define aes_gcm_enc_update_vaes_avx2 ring_core_0_17_14__aes_gcm_enc_update_vaes_avx2
#define aes_hw_ctr32_encrypt_blocks ring_core_0_17_14__aes_hw_ctr32_encrypt_blocks
#define aes_hw_set_encrypt_key ring_core_0_17_14__aes_hw_set_encrypt_key
#define aes_hw_set_encrypt_key_alt ring_core_0_17_14__aes_hw_set_encrypt_key_alt
#define aes_hw_set_encrypt_key_base ring_core_0_17_14__aes_hw_set_encrypt_key_base
#define aes_nohw_ctr32_encrypt_blocks ring_core_0_17_14__aes_nohw_ctr32_encrypt_blocks
#define aes_nohw_encrypt ring_core_0_17_14__aes_nohw_encrypt
#define aes_nohw_set_encrypt_key ring_core_0_17_14__aes_nohw_set_encrypt_key
#define aesni_gcm_decrypt ring_core_0_17_14__aesni_gcm_decrypt
#define aesni_gcm_encrypt ring_core_0_17_14__aesni_gcm_encrypt
#define bn_from_montgomery_in_place ring_core_0_17_14__bn_from_montgomery_in_place
#define bn_gather5 ring_core_0_17_14__bn_gather5
#define bn_mul_mont ring_core_0_17_14__bn_mul_mont
#define bn_mul_mont_nohw ring_core_0_17_14__bn_mul_mont_nohw
#define bn_mul4x_mont ring_core_0_17_14__bn_mul4x_mont
#define bn_mulx4x_mont ring_core_0_17_14__bn_mulx4x_mont
#define bn_mul8x_mont_neon ring_core_0_17_14__bn_mul8x_mont_neon
#define bn_mul4x_mont_gather5 ring_core_0_17_14__bn_mul4x_mont_gather5
#define bn_mulx4x_mont_gather5 ring_core_0_17_14__bn_mulx4x_mont_gather5
#define bn_neg_inv_mod_r_u64 ring_core_0_17_14__bn_neg_inv_mod_r_u64
#define bn_power5_nohw ring_core_0_17_14__bn_power5_nohw
#define bn_powerx5 ring_core_0_17_14__bn_powerx5
#define bn_scatter5 ring_core_0_17_14__bn_scatter5
#define bn_sqr8x_internal ring_core_0_17_14__bn_sqr8x_internal
#define bn_sqr8x_mont ring_core_0_17_14__bn_sqr8x_mont
#define bn_sqrx8x_internal ring_core_0_17_14__bn_sqrx8x_internal
#define bsaes_ctr32_encrypt_blocks ring_core_0_17_14__bsaes_ctr32_encrypt_blocks
#define bssl_constant_time_test_conditional_memcpy ring_core_0_17_14__bssl_constant_time_test_conditional_memcpy
#define bssl_constant_time_test_conditional_memxor ring_core_0_17_14__bssl_constant_time_test_conditional_memxor
#define bssl_constant_time_test_main ring_core_0_17_14__bssl_constant_time_test_main
#define chacha20_poly1305_open ring_core_0_17_14__chacha20_poly1305_open
#define chacha20_poly1305_open_avx2 ring_core_0_17_14__chacha20_poly1305_open_avx2
#define chacha20_poly1305_open_sse41 ring_core_0_17_14__chacha20_poly1305_open_sse41
#define chacha20_poly1305_seal ring_core_0_17_14__chacha20_poly1305_seal
#define chacha20_poly1305_seal_avx2 ring_core_0_17_14__chacha20_poly1305_seal_avx2
#define chacha20_poly1305_seal_sse41 ring_core_0_17_14__chacha20_poly1305_seal_sse41
#define ecp_nistz256_mul_mont_adx ring_core_0_17_14__ecp_nistz256_mul_mont_adx
#define ecp_nistz256_mul_mont_nohw ring_core_0_17_14__ecp_nistz256_mul_mont_nohw
#define ecp_nistz256_ord_mul_mont_adx ring_core_0_17_14__ecp_nistz256_ord_mul_mont_adx
#define ecp_nistz256_ord_mul_mont_nohw ring_core_0_17_14__ecp_nistz256_ord_mul_mont_nohw
#define ecp_nistz256_ord_sqr_mont_adx ring_core_0_17_14__ecp_nistz256_ord_sqr_mont_adx
#define ecp_nistz256_ord_sqr_mont_nohw ring_core_0_17_14__ecp_nistz256_ord_sqr_mont_nohw
#define ecp_nistz256_point_add_adx ring_core_0_17_14__ecp_nistz256_point_add_adx
#define ecp_nistz256_point_add_nohw ring_core_0_17_14__ecp_nistz256_point_add_nohw
#define ecp_nistz256_point_add_affine_adx ring_core_0_17_14__ecp_nistz256_point_add_affine_adx
#define ecp_nistz256_point_add_affine_nohw ring_core_0_17_14__ecp_nistz256_point_add_affine_nohw
#define ecp_nistz256_point_double_adx ring_core_0_17_14__ecp_nistz256_point_double_adx
#define ecp_nistz256_point_double_nohw ring_core_0_17_14__ecp_nistz256_point_double_nohw
#define ecp_nistz256_select_w5_avx2 ring_core_0_17_14__ecp_nistz256_select_w5_avx2
#define ecp_nistz256_select_w5_nohw ring_core_0_17_14__ecp_nistz256_select_w5_nohw
#define ecp_nistz256_select_w7_avx2 ring_core_0_17_14__ecp_nistz256_select_w7_avx2
#define ecp_nistz256_select_w7_nohw ring_core_0_17_14__ecp_nistz256_select_w7_nohw
#define ecp_nistz256_sqr_mont_adx ring_core_0_17_14__ecp_nistz256_sqr_mont_adx
#define ecp_nistz256_sqr_mont_nohw ring_core_0_17_14__ecp_nistz256_sqr_mont_nohw
#define fiat_curve25519_adx_mul ring_core_0_17_14__fiat_curve25519_adx_mul
#define fiat_curve25519_adx_square ring_core_0_17_14__fiat_curve25519_adx_square
#define gcm_ghash_avx ring_core_0_17_14__gcm_ghash_avx
#define gcm_ghash_clmul ring_core_0_17_14__gcm_ghash_clmul
#define gcm_ghash_neon ring_core_0_17_14__gcm_ghash_neon
#define gcm_ghash_vpclmulqdq_avx2_1 ring_core_0_17_14__gcm_ghash_vpclmulqdq_avx2_1
#define gcm_gmult_clmul ring_core_0_17_14__gcm_gmult_clmul
#define gcm_gmult_neon ring_core_0_17_14__gcm_gmult_neon
#define gcm_init_avx ring_core_0_17_14__gcm_init_avx
#define gcm_init_clmul ring_core_0_17_14__gcm_init_clmul
#define gcm_init_neon ring_core_0_17_14__gcm_init_neon
#define gcm_init_vpclmulqdq_avx2 ring_core_0_17_14__gcm_init_vpclmulqdq_avx2
#define k25519Precomp ring_core_0_17_14__k25519Precomp
#define limbs_mul_add_limb ring_core_0_17_14__limbs_mul_add_limb
#define little_endian_bytes_from_scalar ring_core_0_17_14__little_endian_bytes_from_scalar
#define ecp_nistz256_neg ring_core_0_17_14__ecp_nistz256_neg
#define ecp_nistz256_select_w5 ring_core_0_17_14__ecp_nistz256_select_w5
#define ecp_nistz256_select_w7 ring_core_0_17_14__ecp_nistz256_select_w7
#define neon_available ring_core_0_17_14__neon_available
#define p256_mul_mont ring_core_0_17_14__p256_mul_mont
#define p256_point_add ring_core_0_17_14__p256_point_add
#define p256_point_add_affine ring_core_0_17_14__p256_point_add_affine
#define p256_point_double ring_core_0_17_14__p256_point_double
#define p256_point_mul ring_core_0_17_14__p256_point_mul
#define p256_point_mul_base ring_core_0_17_14__p256_point_mul_base
#define p256_point_mul_base_vartime ring_core_0_17_14__p256_point_mul_base_vartime
#define p256_scalar_mul_mont ring_core_0_17_14__p256_scalar_mul_mont
#define p256_scalar_sqr_rep_mont ring_core_0_17_14__p256_scalar_sqr_rep_mont
#define p256_sqr_mont ring_core_0_17_14__p256_sqr_mont
#define p384_elem_div_by_2 ring_core_0_17_14__p384_elem_div_by_2
#define p384_elem_mul_mont ring_core_0_17_14__p384_elem_mul_mont
#define p384_elem_neg ring_core_0_17_14__p384_elem_neg
#define p384_elem_sub ring_core_0_17_14__p384_elem_sub
#define p384_point_add ring_core_0_17_14__p384_point_add
#define p384_point_double ring_core_0_17_14__p384_point_double
#define p384_point_mul ring_core_0_17_14__p384_point_mul
#define p384_scalar_mul_mont ring_core_0_17_14__p384_scalar_mul_mont
#define openssl_poly1305_neon2_addmulmod ring_core_0_17_14__openssl_poly1305_neon2_addmulmod
#define openssl_poly1305_neon2_blocks ring_core_0_17_14__openssl_poly1305_neon2_blocks
#define sha256_block_data_order ring_core_0_17_14__sha256_block_data_order
#define sha256_block_data_order_avx ring_core_0_17_14__sha256_block_data_order_avx
#define sha256_block_data_order_ssse3 ring_core_0_17_14__sha256_block_data_order_ssse3
#define sha256_block_data_order_hw ring_core_0_17_14__sha256_block_data_order_hw
#define sha256_block_data_order_neon ring_core_0_17_14__sha256_block_data_order_neon
#define sha256_block_data_order_nohw ring_core_0_17_14__sha256_block_data_order_nohw
#define sha512_block_data_order ring_core_0_17_14__sha512_block_data_order
#define sha512_block_data_order_avx ring_core_0_17_14__sha512_block_data_order_avx
#define sha512_block_data_order_hw ring_core_0_17_14__sha512_block_data_order_hw
#define sha512_block_data_order_neon ring_core_0_17_14__sha512_block_data_order_neon
#define sha512_block_data_order_nohw ring_core_0_17_14__sha512_block_data_order_nohw
#define vpaes_ctr32_encrypt_blocks ring_core_0_17_14__vpaes_ctr32_encrypt_blocks
#define vpaes_encrypt ring_core_0_17_14__vpaes_encrypt
#define vpaes_encrypt_key_to_bsaes ring_core_0_17_14__vpaes_encrypt_key_to_bsaes
#define vpaes_set_encrypt_key ring_core_0_17_14__vpaes_set_encrypt_key
#define x25519_NEON ring_core_0_17_14__x25519_NEON
#define x25519_fe_invert ring_core_0_17_14__x25519_fe_invert
#define x25519_fe_isnegative ring_core_0_17_14__x25519_fe_isnegative
#define x25519_fe_mul_ttt ring_core_0_17_14__x25519_fe_mul_ttt
#define x25519_fe_neg ring_core_0_17_14__x25519_fe_neg
#define x25519_fe_tobytes ring_core_0_17_14__x25519_fe_tobytes
#define x25519_ge_double_scalarmult_vartime ring_core_0_17_14__x25519_ge_double_scalarmult_vartime
#define x25519_ge_frombytes_vartime ring_core_0_17_14__x25519_ge_frombytes_vartime
#define x25519_ge_scalarmult_base ring_core_0_17_14__x25519_ge_scalarmult_base
#define x25519_ge_scalarmult_base_adx ring_core_0_17_14__x25519_ge_scalarmult_base_adx
#define x25519_public_from_private_generic_masked ring_core_0_17_14__x25519_public_from_private_generic_masked
#define x25519_sc_mask ring_core_0_17_14__x25519_sc_mask
#define x25519_sc_muladd ring_core_0_17_14__x25519_sc_muladd
#define x25519_sc_reduce ring_core_0_17_14__x25519_sc_reduce
#define x25519_scalar_mult_adx ring_core_0_17_14__x25519_scalar_mult_adx
#define x25519_scalar_mult_generic_masked ring_core_0_17_14__x25519_scalar_mult_generic_masked

#endif
