// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::ffi;
use glib::GStr;

#[cfg(feature = "v2_72")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_72")))]
#[doc(alias = "G_DEBUG_CONTROLLER_EXTENSION_POINT_NAME")]
pub static DEBUG_CONTROLLER_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_DEBUG_CONTROLLER_EXTENSION_POINT_NAME) };
#[cfg(feature = "v2_58")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_58")))]
#[doc(alias = "G_DRIVE_IDENTIFIER_KIND_UNIX_DEVICE")]
pub static DRIVE_IDENTIFIER_KIND_UNIX_DEVICE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_DRIVE_IDENTIFIER_KIND_UNIX_DEVICE) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_DELETE")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_DELETE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_DELETE) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_READ")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_READ: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_READ) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_RENAME")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_RENAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_RENAME) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_TRASH")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_TRASH: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_TRASH) };
#[doc(alias = "G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE")]
pub static FILE_ATTRIBUTE_ACCESS_CAN_WRITE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE) };
#[doc(alias = "G_FILE_ATTRIBUTE_DOS_IS_ARCHIVE")]
pub static FILE_ATTRIBUTE_DOS_IS_ARCHIVE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_DOS_IS_ARCHIVE) };
#[cfg(feature = "v2_60")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_60")))]
#[doc(alias = "G_FILE_ATTRIBUTE_DOS_IS_MOUNTPOINT")]
pub static FILE_ATTRIBUTE_DOS_IS_MOUNTPOINT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_DOS_IS_MOUNTPOINT) };
#[doc(alias = "G_FILE_ATTRIBUTE_DOS_IS_SYSTEM")]
pub static FILE_ATTRIBUTE_DOS_IS_SYSTEM: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_DOS_IS_SYSTEM) };
#[cfg(feature = "v2_60")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_60")))]
#[doc(alias = "G_FILE_ATTRIBUTE_DOS_REPARSE_POINT_TAG")]
pub static FILE_ATTRIBUTE_DOS_REPARSE_POINT_TAG: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_DOS_REPARSE_POINT_TAG) };
#[doc(alias = "G_FILE_ATTRIBUTE_ETAG_VALUE")]
pub static FILE_ATTRIBUTE_ETAG_VALUE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ETAG_VALUE) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_FREE")]
pub static FILE_ATTRIBUTE_FILESYSTEM_FREE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_FREE) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_READONLY")]
pub static FILE_ATTRIBUTE_FILESYSTEM_READONLY: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_READONLY) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_REMOTE")]
pub static FILE_ATTRIBUTE_FILESYSTEM_REMOTE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_REMOTE) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_SIZE")]
pub static FILE_ATTRIBUTE_FILESYSTEM_SIZE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_SIZE) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_TYPE")]
pub static FILE_ATTRIBUTE_FILESYSTEM_TYPE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_TYPE) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_USED")]
pub static FILE_ATTRIBUTE_FILESYSTEM_USED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_USED) };
#[doc(alias = "G_FILE_ATTRIBUTE_FILESYSTEM_USE_PREVIEW")]
pub static FILE_ATTRIBUTE_FILESYSTEM_USE_PREVIEW: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_FILESYSTEM_USE_PREVIEW) };
#[doc(alias = "G_FILE_ATTRIBUTE_GVFS_BACKEND")]
pub static FILE_ATTRIBUTE_GVFS_BACKEND: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_GVFS_BACKEND) };
#[doc(alias = "G_FILE_ATTRIBUTE_ID_FILE")]
pub static FILE_ATTRIBUTE_ID_FILE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ID_FILE) };
#[doc(alias = "G_FILE_ATTRIBUTE_ID_FILESYSTEM")]
pub static FILE_ATTRIBUTE_ID_FILESYSTEM: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_ID_FILESYSTEM) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_EJECT")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_EJECT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_EJECT) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_MOUNT) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_POLL")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_POLL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_POLL) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_START")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_START: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_START) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_START_DEGRADED")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_START_DEGRADED: &GStr = unsafe {
    GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_START_DEGRADED)
};
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_STOP")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_STOP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_STOP) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_CAN_UNMOUNT")]
pub static FILE_ATTRIBUTE_MOUNTABLE_CAN_UNMOUNT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_CAN_UNMOUNT) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_HAL_UDI")]
pub static FILE_ATTRIBUTE_MOUNTABLE_HAL_UDI: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_HAL_UDI) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_IS_MEDIA_CHECK_AUTOMATIC")]
pub static FILE_ATTRIBUTE_MOUNTABLE_IS_MEDIA_CHECK_AUTOMATIC: &GStr = unsafe {
    GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_IS_MEDIA_CHECK_AUTOMATIC)
};
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_START_STOP_TYPE")]
pub static FILE_ATTRIBUTE_MOUNTABLE_START_STOP_TYPE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_START_STOP_TYPE) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE")]
pub static FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE) };
#[doc(alias = "G_FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE_FILE")]
pub static FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE_FILE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_MOUNTABLE_UNIX_DEVICE_FILE) };
#[doc(alias = "G_FILE_ATTRIBUTE_OWNER_GROUP")]
pub static FILE_ATTRIBUTE_OWNER_GROUP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_OWNER_GROUP) };
#[doc(alias = "G_FILE_ATTRIBUTE_OWNER_USER")]
pub static FILE_ATTRIBUTE_OWNER_USER: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_OWNER_USER) };
#[doc(alias = "G_FILE_ATTRIBUTE_OWNER_USER_REAL")]
pub static FILE_ATTRIBUTE_OWNER_USER_REAL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_OWNER_USER_REAL) };
#[doc(alias = "G_FILE_ATTRIBUTE_PREVIEW_ICON")]
pub static FILE_ATTRIBUTE_PREVIEW_ICON: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_PREVIEW_ICON) };
#[doc(alias = "G_FILE_ATTRIBUTE_RECENT_MODIFIED")]
pub static FILE_ATTRIBUTE_RECENT_MODIFIED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_RECENT_MODIFIED) };
#[doc(alias = "G_FILE_ATTRIBUTE_SELINUX_CONTEXT")]
pub static FILE_ATTRIBUTE_SELINUX_CONTEXT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_SELINUX_CONTEXT) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_ALLOCATED_SIZE")]
pub static FILE_ATTRIBUTE_STANDARD_ALLOCATED_SIZE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_ALLOCATED_SIZE) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE")]
pub static FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_COPY_NAME")]
pub static FILE_ATTRIBUTE_STANDARD_COPY_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_COPY_NAME) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_DESCRIPTION")]
pub static FILE_ATTRIBUTE_STANDARD_DESCRIPTION: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_DESCRIPTION) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME")]
pub static FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME")]
pub static FILE_ATTRIBUTE_STANDARD_EDIT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE")]
pub static FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_FAST_CONTENT_TYPE) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_ICON")]
pub static FILE_ATTRIBUTE_STANDARD_ICON: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_ICON) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_IS_BACKUP")]
pub static FILE_ATTRIBUTE_STANDARD_IS_BACKUP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_IS_BACKUP) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_IS_HIDDEN")]
pub static FILE_ATTRIBUTE_STANDARD_IS_HIDDEN: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_IS_HIDDEN) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK")]
pub static FILE_ATTRIBUTE_STANDARD_IS_SYMLINK: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_IS_VIRTUAL")]
pub static FILE_ATTRIBUTE_STANDARD_IS_VIRTUAL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_IS_VIRTUAL) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_IS_VOLATILE")]
pub static FILE_ATTRIBUTE_STANDARD_IS_VOLATILE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_IS_VOLATILE) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_NAME")]
pub static FILE_ATTRIBUTE_STANDARD_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_NAME) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_SIZE")]
pub static FILE_ATTRIBUTE_STANDARD_SIZE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_SIZE) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_SORT_ORDER")]
pub static FILE_ATTRIBUTE_STANDARD_SORT_ORDER: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_SORT_ORDER) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_SYMBOLIC_ICON")]
pub static FILE_ATTRIBUTE_STANDARD_SYMBOLIC_ICON: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_SYMBOLIC_ICON) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET")]
pub static FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_TARGET_URI")]
pub static FILE_ATTRIBUTE_STANDARD_TARGET_URI: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_TARGET_URI) };
#[doc(alias = "G_FILE_ATTRIBUTE_STANDARD_TYPE")]
pub static FILE_ATTRIBUTE_STANDARD_TYPE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_STANDARD_TYPE) };
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAILING_FAILED")]
pub static FILE_ATTRIBUTE_THUMBNAILING_FAILED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAILING_FAILED) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_LARGE")]
pub static FILE_ATTRIBUTE_THUMBNAILING_FAILED_LARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_LARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_NORMAL")]
pub static FILE_ATTRIBUTE_THUMBNAILING_FAILED_NORMAL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_NORMAL) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_XLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAILING_FAILED_XLARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_XLARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_XXLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAILING_FAILED_XXLARGE: &GStr = unsafe {
    GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAILING_FAILED_XXLARGE)
};
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID")]
pub static FILE_ATTRIBUTE_THUMBNAIL_IS_VALID: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_LARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_LARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_LARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_NORMAL")]
pub static FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_NORMAL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_NORMAL) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XLARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XLARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XXLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XXLARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_IS_VALID_XXLARGE) };
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_PATH")]
pub static FILE_ATTRIBUTE_THUMBNAIL_PATH: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_PATH) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_PATH_LARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_PATH_LARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_PATH_LARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_PATH_NORMAL")]
pub static FILE_ATTRIBUTE_THUMBNAIL_PATH_NORMAL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_PATH_NORMAL) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_PATH_XLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_PATH_XLARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_PATH_XLARGE) };
#[cfg(feature = "v2_76")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_76")))]
#[doc(alias = "G_FILE_ATTRIBUTE_THUMBNAIL_PATH_XXLARGE")]
pub static FILE_ATTRIBUTE_THUMBNAIL_PATH_XXLARGE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_THUMBNAIL_PATH_XXLARGE) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_ACCESS")]
pub static FILE_ATTRIBUTE_TIME_ACCESS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_ACCESS) };
#[cfg(feature = "v2_74")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_74")))]
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_ACCESS_NSEC")]
pub static FILE_ATTRIBUTE_TIME_ACCESS_NSEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_ACCESS_NSEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_ACCESS_USEC")]
pub static FILE_ATTRIBUTE_TIME_ACCESS_USEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_ACCESS_USEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CHANGED")]
pub static FILE_ATTRIBUTE_TIME_CHANGED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CHANGED) };
#[cfg(feature = "v2_74")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_74")))]
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CHANGED_NSEC")]
pub static FILE_ATTRIBUTE_TIME_CHANGED_NSEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CHANGED_NSEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CHANGED_USEC")]
pub static FILE_ATTRIBUTE_TIME_CHANGED_USEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CHANGED_USEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CREATED")]
pub static FILE_ATTRIBUTE_TIME_CREATED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CREATED) };
#[cfg(feature = "v2_74")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_74")))]
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CREATED_NSEC")]
pub static FILE_ATTRIBUTE_TIME_CREATED_NSEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CREATED_NSEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_CREATED_USEC")]
pub static FILE_ATTRIBUTE_TIME_CREATED_USEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_CREATED_USEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_MODIFIED")]
pub static FILE_ATTRIBUTE_TIME_MODIFIED: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_MODIFIED) };
#[cfg(feature = "v2_74")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_74")))]
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_MODIFIED_NSEC")]
pub static FILE_ATTRIBUTE_TIME_MODIFIED_NSEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_MODIFIED_NSEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC")]
pub static FILE_ATTRIBUTE_TIME_MODIFIED_USEC: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC) };
#[doc(alias = "G_FILE_ATTRIBUTE_TRASH_DELETION_DATE")]
pub static FILE_ATTRIBUTE_TRASH_DELETION_DATE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TRASH_DELETION_DATE) };
#[doc(alias = "G_FILE_ATTRIBUTE_TRASH_ITEM_COUNT")]
pub static FILE_ATTRIBUTE_TRASH_ITEM_COUNT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TRASH_ITEM_COUNT) };
#[doc(alias = "G_FILE_ATTRIBUTE_TRASH_ORIG_PATH")]
pub static FILE_ATTRIBUTE_TRASH_ORIG_PATH: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_TRASH_ORIG_PATH) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_BLOCKS")]
pub static FILE_ATTRIBUTE_UNIX_BLOCKS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_BLOCKS) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE")]
pub static FILE_ATTRIBUTE_UNIX_BLOCK_SIZE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_DEVICE")]
pub static FILE_ATTRIBUTE_UNIX_DEVICE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_DEVICE) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_GID")]
pub static FILE_ATTRIBUTE_UNIX_GID: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_GID) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_INODE")]
pub static FILE_ATTRIBUTE_UNIX_INODE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_INODE) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_IS_MOUNTPOINT")]
pub static FILE_ATTRIBUTE_UNIX_IS_MOUNTPOINT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_IS_MOUNTPOINT) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_MODE")]
pub static FILE_ATTRIBUTE_UNIX_MODE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_MODE) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_NLINK")]
pub static FILE_ATTRIBUTE_UNIX_NLINK: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_NLINK) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_RDEV")]
pub static FILE_ATTRIBUTE_UNIX_RDEV: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_RDEV) };
#[doc(alias = "G_FILE_ATTRIBUTE_UNIX_UID")]
pub static FILE_ATTRIBUTE_UNIX_UID: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_FILE_ATTRIBUTE_UNIX_UID) };
#[cfg(feature = "v2_64")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_64")))]
#[doc(alias = "G_MEMORY_MONITOR_EXTENSION_POINT_NAME")]
pub static MEMORY_MONITOR_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MEMORY_MONITOR_EXTENSION_POINT_NAME) };
#[doc(alias = "G_MENU_ATTRIBUTE_ACTION")]
pub static MENU_ATTRIBUTE_ACTION: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_ATTRIBUTE_ACTION) };
#[doc(alias = "G_MENU_ATTRIBUTE_ACTION_NAMESPACE")]
pub static MENU_ATTRIBUTE_ACTION_NAMESPACE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_ATTRIBUTE_ACTION_NAMESPACE) };
#[doc(alias = "G_MENU_ATTRIBUTE_ICON")]
pub static MENU_ATTRIBUTE_ICON: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_ATTRIBUTE_ICON) };
#[doc(alias = "G_MENU_ATTRIBUTE_LABEL")]
pub static MENU_ATTRIBUTE_LABEL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_ATTRIBUTE_LABEL) };
#[doc(alias = "G_MENU_ATTRIBUTE_TARGET")]
pub static MENU_ATTRIBUTE_TARGET: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_ATTRIBUTE_TARGET) };
#[doc(alias = "G_MENU_LINK_SECTION")]
pub static MENU_LINK_SECTION: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_LINK_SECTION) };
#[doc(alias = "G_MENU_LINK_SUBMENU")]
pub static MENU_LINK_SUBMENU: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_MENU_LINK_SUBMENU) };
#[doc(alias = "G_NATIVE_VOLUME_MONITOR_EXTENSION_POINT_NAME")]
pub static NATIVE_VOLUME_MONITOR_EXTENSION_POINT_NAME: &GStr = unsafe {
    GStr::from_utf8_with_nul_unchecked(ffi::G_NATIVE_VOLUME_MONITOR_EXTENSION_POINT_NAME)
};
#[doc(alias = "G_NETWORK_MONITOR_EXTENSION_POINT_NAME")]
pub static NETWORK_MONITOR_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_NETWORK_MONITOR_EXTENSION_POINT_NAME) };
#[cfg(feature = "v2_70")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_70")))]
#[doc(alias = "G_POWER_PROFILE_MONITOR_EXTENSION_POINT_NAME")]
pub static POWER_PROFILE_MONITOR_EXTENSION_POINT_NAME: &GStr = unsafe {
    GStr::from_utf8_with_nul_unchecked(ffi::G_POWER_PROFILE_MONITOR_EXTENSION_POINT_NAME)
};
#[doc(alias = "G_PROXY_EXTENSION_POINT_NAME")]
pub static PROXY_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_PROXY_EXTENSION_POINT_NAME) };
#[doc(alias = "G_PROXY_RESOLVER_EXTENSION_POINT_NAME")]
pub static PROXY_RESOLVER_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_PROXY_RESOLVER_EXTENSION_POINT_NAME) };
#[doc(alias = "G_SETTINGS_BACKEND_EXTENSION_POINT_NAME")]
pub static SETTINGS_BACKEND_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_SETTINGS_BACKEND_EXTENSION_POINT_NAME) };
#[doc(alias = "G_TLS_BACKEND_EXTENSION_POINT_NAME")]
pub static TLS_BACKEND_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_TLS_BACKEND_EXTENSION_POINT_NAME) };
#[doc(alias = "G_TLS_DATABASE_PURPOSE_AUTHENTICATE_CLIENT")]
pub static TLS_DATABASE_PURPOSE_AUTHENTICATE_CLIENT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_TLS_DATABASE_PURPOSE_AUTHENTICATE_CLIENT) };
#[doc(alias = "G_TLS_DATABASE_PURPOSE_AUTHENTICATE_SERVER")]
pub static TLS_DATABASE_PURPOSE_AUTHENTICATE_SERVER: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_TLS_DATABASE_PURPOSE_AUTHENTICATE_SERVER) };
#[doc(alias = "G_VFS_EXTENSION_POINT_NAME")]
pub static VFS_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VFS_EXTENSION_POINT_NAME) };
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_CLASS")]
pub static VOLUME_IDENTIFIER_KIND_CLASS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_CLASS) };
#[cfg_attr(feature = "v2_58", deprecated = "Since 2.58")]
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_HAL_UDI")]
pub static VOLUME_IDENTIFIER_KIND_HAL_UDI: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_HAL_UDI) };
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_LABEL")]
pub static VOLUME_IDENTIFIER_KIND_LABEL: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_LABEL) };
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_NFS_MOUNT")]
pub static VOLUME_IDENTIFIER_KIND_NFS_MOUNT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_NFS_MOUNT) };
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE")]
pub static VOLUME_IDENTIFIER_KIND_UNIX_DEVICE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE) };
#[doc(alias = "G_VOLUME_IDENTIFIER_KIND_UUID")]
pub static VOLUME_IDENTIFIER_KIND_UUID: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_IDENTIFIER_KIND_UUID) };
#[doc(alias = "G_VOLUME_MONITOR_EXTENSION_POINT_NAME")]
pub static VOLUME_MONITOR_EXTENSION_POINT_NAME: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::G_VOLUME_MONITOR_EXTENSION_POINT_NAME) };
